@@ -1,5 +1,5 @@
 __author__ = 'Robbert Harms'
 __date__ = '2018-08-01'
 __maintainer__ = 'Robbert Harms'
-__email__ = 'robbert.harms@maastrichtuniversity.nl'
+__email__ = 'robbert@xkls.nl'
 __licence__ = 'LGPL v3'
