@@ -3,21 +3,87 @@ import logging
 import multiprocessing
 import numbers
 import os
+from collections import Iterable, Mapping, OrderedDict
 from contextlib import contextmanager
 from functools import reduce
-
 import numpy as np
 import pyopencl as cl
 import pyopencl.array as cl_array
 import tatsu
-
-from mot.lib.cl_data_type import SimpleCLDataType
+from pkg_resources import resource_filename
 
 __author__ = 'Robbert Harms'
 __date__ = "2014-05-13"
 __license__ = "LGPL v3"
 __maintainer__ = "Robbert Harms"
-__email__ = "robbert.harms@maastrichtuniversity.nl"
+__email__ = "robbert@xkls.nl"
+
+
+def convert_inputs_to_kernel_data(inputs, cl_parameters, nmr_instances):
+    """Convert a list or dictionary of input values to an appropriate :class:`~mot.lib.kernel_data.KernelData` object.
+
+    Args:
+        inputs (Iterable[Union(ndarray, mot.lib.utils.KernelData)]
+                    or Mapping[str: Union(ndarray, mot.lib.utils.KernelData)]):
+            for each CL parameter matching input data. This function accepts either an iterable with one value
+            per parameter, or a mapping with for every parameter a corresponding value. The elements can either
+            be scalars, numpy ndarrays  or KernelData objects. If a scalar is given we load it as is. If an ndarray
+            is given we will load it read/write by default.
+        cl_parameters (List[mot.lib.cl_function.CLFunctionParameter]): list of function parameters corresponding to the
+            function we plan on using the kernel data for.
+        nmr_instances (int): the number of parallel processes supported by the data
+
+    Returns:
+        OrderedDict[str: mot.lib.kernel_data.KernelData]: mapping parameter names to kernel data objects.
+    """
+    from mot.lib.kernel_data import KernelData, Scalar, Array
+
+    if isinstance(inputs, Iterable) and not isinstance(inputs, Mapping):
+        inputs = list(inputs)
+        if len(inputs) != len(cl_parameters):
+            raise ValueError('The length of the input list ({}), does not equal '
+                             'the number of parameters ({})'.format(len(inputs), len(cl_parameters)))
+
+        param_names = [param.name for param in cl_parameters]
+        inputs = dict(zip(param_names, inputs))
+
+    for param in cl_parameters:
+        if param.name not in inputs:
+            names = [param.name for param in cl_parameters]
+            missing_names = set(names) - set(inputs.keys())
+            raise ValueError('Some parameters are missing an input value, '
+                             'required parameters are: {}, given inputs are: {}, missing are: {}'.format(
+                names, list(inputs.keys()), list(missing_names)))
+
+    def get_data_object(param):
+        input_data = inputs[param.name]
+
+        if input_data is None:
+            return Scalar(0)
+        elif isinstance(input_data, KernelData):
+            return input_data
+
+        input_data = np.array(input_data)
+
+        if param.is_vector_type and np.squeeze(input_data).shape[0] == 3:
+            return Scalar(inputs[param.name], ctype=param.ctype)
+        elif is_scalar(inputs[param.name]) and not (param.is_pointer_type or param.is_array_type):
+            return Scalar(inputs[param.name])
+        else:
+            if is_scalar(inputs[param.name]):
+                if param.is_pointer_type or param.is_array_type:
+                    data = np.ones(nmr_instances) * inputs[param.name]
+                    return Array(data, ctype=param.ctype, mode='rw')
+                else:
+                    return Scalar(inputs[param.name])
+            else:
+                data = inputs[param.name]
+                if param.is_pointer_type or param.is_array_type:
+                    return Array(data, ctype=param.ctype, mode='rw')
+                else:
+                    return Array(data, ctype=param.ctype, mode='r', as_scalar=True)
+
+    return OrderedDict([(param.name.replace('.', '_'), get_data_object(param)) for param in cl_parameters])
 
 
 def add_include_guards(cl_str, guard_name=None):
@@ -69,16 +135,21 @@ def ctype_to_dtype(cl_type, mot_float_type='float'):
     Returns:
         dtype: the numpy datatype
     """
-    data_type = SimpleCLDataType.from_string(cl_type)
+    if is_vector_ctype(cl_type):
+        raw_type, vector_length = split_vector_ctype(cl_type)
 
-    if data_type.is_vector_type:
-        if data_type.raw_data_type.startswith('mot_float_type'):
-            data_type = SimpleCLDataType.from_string(mot_float_type + str(data_type.vector_length))
-        vector_type = data_type.raw_data_type + str(data_type.vector_length)
+        if raw_type == 'mot_float_type':
+            if is_vector_ctype(mot_float_type):
+                raw_type, _ = split_vector_ctype(mot_float_type)
+            else:
+                raw_type = mot_float_type
+
+        vector_type = raw_type + str(vector_length)
         return getattr(cl_array.vec, vector_type)
     else:
-        if data_type.raw_data_type.startswith('mot_float_type'):
-            data_type = SimpleCLDataType.from_string(mot_float_type)
+        if cl_type == 'mot_float_type':
+            cl_type = mot_float_type
+
         data_types = [
             ('char', np.int8),
             ('uchar', np.uint8),
@@ -92,7 +163,7 @@ def ctype_to_dtype(cl_type, mot_float_type='float'):
             ('double', np.float64),
         ]
         for ctype, dtype in data_types:
-            if ctype == data_type.raw_data_type:
+            if ctype == cl_type:
                 return dtype
 
 
@@ -101,26 +172,28 @@ def convert_data_to_dtype(data, data_type, mot_float_type='float'):
 
     Args:
         data (ndarray): The value to convert to the correct numpy type
-        data_type (str or mot.lib.cl_data_type.CLDataType): the data type we need to convert the data to
-        mot_float_type (str or mot.lib.cl_data_type.CLDataType): the data type of the current ``mot_float_type``
+        data_type (str): the data type we need to convert the data to
+        mot_float_type (str): the data type of the current ``mot_float_type``
 
     Returns:
         ndarray: the input data but then converted to the desired numpy data type
     """
-    if isinstance(data_type, str):
-        data_type = SimpleCLDataType.from_string(data_type)
-    if isinstance(mot_float_type, str):
-        mot_float_type = SimpleCLDataType.from_string(mot_float_type)
-
-    scalar_dtype = ctype_to_dtype(data_type.raw_data_type, mot_float_type.raw_data_type)
+    scalar_dtype = ctype_to_dtype(data_type, mot_float_type)
 
     if isinstance(data, numbers.Number):
         data = scalar_dtype(data)
 
-    if data_type.is_vector_type:
+    if is_vector_ctype(data_type):
         shape = data.shape
-        dtype = ctype_to_dtype(data_type, mot_float_type.raw_data_type)
-        ve = np.zeros(shape[:-1], dtype=dtype)
+        dtype = ctype_to_dtype(data_type, mot_float_type)
+
+        if len(shape) > 1:
+            ve = np.zeros(shape[:-1], dtype=dtype)
+        else:
+            ve = np.zeros(1, dtype=dtype)
+
+        if data.dtype == dtype:
+            return data
 
         if len(shape) == 1:
             for vector_ind in range(shape[0]):
@@ -135,8 +208,39 @@ def convert_data_to_dtype(data, data_type, mot_float_type='float'):
                     for vector_ind in range(data.shape[2]):
                         ve[i, j][vector_ind] = data[i, j, vector_ind]
 
-        return np.require(ve, requirements=['C', 'A', 'O'])
-    return np.require(data, scalar_dtype, ['C', 'A', 'O'])
+        return ve
+    return np.require(data, scalar_dtype)
+
+
+def split_vector_ctype(ctype):
+    """Split a vector ctype into a raw ctype and the vector length.
+
+    If the given ctype is not a vector type, we raise an error. I
+
+    Args:
+         ctype (str): the ctype to possibly split into a raw ctype and the vector length
+
+    Returns:
+        tuple: the raw ctype and the vector length
+    """
+    if not is_vector_ctype(ctype):
+        raise ValueError('The given ctype is not a vector type.')
+    for vector_length in [2, 3, 4, 8, 16]:
+        if ctype.endswith(str(vector_length)):
+            vector_str_len = len(str(vector_length))
+            return ctype[:-vector_str_len], int(ctype[-vector_str_len:])
+
+
+def is_vector_ctype(ctype):
+    """Test if the given ctype is a vector type. That is, if it ends with 2, 3, 4, 8 or 16.
+
+    Args:
+        ctype (str): the ctype to test if it is an OpenCL vector type
+
+    Returns:
+        bool: if it is a vector type or not
+    """
+    return any(ctype.endswith(str(i)) for i in [2, 3, 4, 8, 16])
 
 
 def device_type_from_string(cl_device_type_str):
@@ -165,21 +269,27 @@ def device_supports_double(cl_device):
     Returns:
         boolean: True if the given cl_device supports double, false otherwise.
     """
-    return cl_device.get_info(cl.device_info.DOUBLE_FP_CONFIG) == 63
+    dev_extensions = cl_device.extensions.strip().split(' ')
+    return 'cl_khr_fp64' in dev_extensions
 
 
-def get_float_type_def(double_precision):
+def get_cl_utility_definitions(double_precision, include_complex=True):
     """Get the model floating point type definition.
 
-    The MOT_INT_CMP_TYPE is meant for the select() function where you need a long in the case of double precision.
-
     Args:
         double_precision (boolean): if True we will use the double type for the mot_float_type type.
             Else, we will use the single precision float type for the mot_float_type type.
+        include_complex (boolean): if we include support for complex numbers
 
     Returns:
         str: defines the mot_float_type types, the epsilon and the MIN and MAX values.
     """
+    if include_complex:
+        with open(os.path.abspath(resource_filename('mot', 'data/opencl/complex.h')), 'r') as f:
+            complex_number_support = f.read()
+    else:
+        complex_number_support = ''
+
     scipy_constants = '''
         #define MACHEP DBL_EPSILON
         #define MAXLOG log(DBL_MAX)
@@ -187,43 +297,67 @@ def get_float_type_def(double_precision):
         #define EULER 0.577215664901532860606512090082402431 /* Euler constant, from Scipy */
     '''
 
+    helper_funcs = '''
+        /**
+         * Utility function used in splitting work over workitems in a workgroup.
+         *
+         * This automatically loads the current local_id and local_size and will return
+         * batch offset and range accordingly.
+         *
+         * Args:
+         *  nmr_items (INPUT): the total number of items that must be processed
+         *  batch_range (OUTPUT): the number of elements the current workitem must process.
+         *
+         * Returns:
+         *  offset: the offset into the number of items, i.e. the start of processing for this workitem.
+         */
+        uint get_workitem_batch(uint nmr_items, uint* batch_range){
+            uint local_id = get_local_id(0);
+            uint workgroup_size = get_local_size(0);
+
+            uint full_runs = nmr_items / workgroup_size;
+            uint rest = nmr_items % workgroup_size;
+            uint offset = full_runs * local_id;
+
+            *batch_range = full_runs + !(local_id / rest);
+            return offset + min(local_id, rest);
+        }
+    '''
+
     if double_precision:
-        return '''            
+        return '''
             #if __OPENCL_VERSION__ <= CL_VERSION_1_1
                 #pragma OPENCL EXTENSION cl_khr_fp64 : enable
             #endif
-            
+
             #define PYOPENCL_DEFINE_CDOUBLE
-            #include <pyopencl-complex.h>
-        
-            #define mot_float_type double
-            #define mot_float_type2 double2
-            #define mot_float_type4 double4
-            #define mot_float_type8 double8
-            #define mot_float_type16 double16
+
+            typedef double mot_float_type;
+            typedef double2 mot_float_type2;
+            typedef double4 mot_float_type4;
+            typedef double8 mot_float_type8;
+            typedef double16 mot_float_type16;
+
             #define MOT_EPSILON DBL_EPSILON
             #define MOT_MIN DBL_MIN
             #define MOT_MAX DBL_MAX
-            #define MOT_INT_CMP_TYPE long
-        ''' + scipy_constants
+        ''' + scipy_constants + complex_number_support + helper_funcs
     else:
         return '''
             #if __OPENCL_VERSION__ <= CL_VERSION_1_1
                 #pragma OPENCL EXTENSION cl_khr_fp64 : enable
             #endif
-            
-            #include <pyopencl-complex.h>
-            
-            #define mot_float_type float
-            #define mot_float_type2 float2
-            #define mot_float_type4 float4
-            #define mot_float_type8 float8
-            #define mot_float_type16 float16
+
+            typedef float mot_float_type;
+            typedef float2 mot_float_type2;
+            typedef float4 mot_float_type4;
+            typedef float8 mot_float_type8;
+            typedef float16 mot_float_type16;
+
             #define MOT_EPSILON FLT_EPSILON
             #define MOT_MIN FLT_MIN
             #define MOT_MAX FLT_MAX
-            #define MOT_INT_CMP_TYPE int
-        ''' + scipy_constants
+        ''' + scipy_constants + complex_number_support + helper_funcs
 
 
 def topological_sort(data):
@@ -425,78 +559,48 @@ def cartesian(arrays, out=None):
     return out
 
 
-def split_in_batches(nmr_elements, max_batch_size):
-    """Split the total number of elements into batches of the specified maximum size.
+def split_in_batches(nmr_elements, max_batch_size=None, nmr_batches=None):
+    """Split the total number of elements into batches.
+
+    This function has two modes of operation. One is to make batches of the specified maximum size, the other
+    is to make a selected number of batches.
 
     Examples::
-        split_in_batches(30, 8) -> [(0, 8), (8, 15), (16, 23), (24, 29)]
+        split_in_batches(30, max_batch_size=8) -> [(0, 8), (8, 15), (16, 23), (24, 29)]
 
         for batch_start, batch_end in split_in_batches(2000, 100):
             array[batch_start:batch_end]
 
+    Args:
+        nmr_elements (int): the total number of elements to partition
+        max_batch_size (int): the maximum size per batch, if set one can not set nmr_batches.
+        nmr_batches (int): the number of batches to create, if set one can not set max_batch_size.
+
     Yields:
         tuple: the start and end point of the next batch
     """
-    offset = 0
-    elements_left = nmr_elements
-    while elements_left > 0:
-        next_batch = (offset, offset + min(elements_left, max_batch_size))
-        yield next_batch
+    if max_batch_size is not None and nmr_batches is not None:
+        raise ValueError('max_batch_size and nmr_batches can not be set at the same time.')
+    if max_batch_size is None and nmr_batches is None:
+        raise ValueError('Either max_batch_size or nmr_batches must be set.')
 
-        batch_size = min(elements_left, max_batch_size)
-        elements_left -= batch_size
-        offset += batch_size
-
-
-def hessian_to_covariance(hessian, output_singularity=False):
-    """Calculate a covariance matrix from a Hessian by inverting the Hessian.
-
-    Mathematically we can calculate the covariance matrix from the Hessian (the Hessian at the Maximum Likelihood
-    Estimator), by a simple matrix inversion. However, round-off errors can make the Hessian singular, making an
-    exact inverse impossible. This method uses an exact inverse if possible with a fall back on a pseudo inverse.
-    If also the pseudo inverse fails, this function returns zeros as covariance for that Hessian.
-
-    Important: Before the matrix inversion it will set NaN's to 0. After the inversion we make the diagonal
-    (representing the variances of each parameter) positive where needed by taking the absolute.
-
-    Args:
-        hessian (ndarray): a matrix of shape (n, p, p) where for n problems we have a matrix of shape (p, p) for
-            p parameters and we take the inverse for every (p, p) matrix.
-        output_singularity (boolean): if set to True, we additionally output a boolean matrix with the location
-            of singular Hessians.
-
-    Returns:
-        ndarray or tuple: if ``output_singularity`` is set to False, only output the inverse of the Hessians as the
-            covariance matrix. If ``output_singularity`` is set to True this function returns a tuple with:
-            (``covariance_matrix``, ``is_singular``).
-    """
-    hessian = np.nan_to_num(hessian)
-
-    if output_singularity:
-        is_singular = np.zeros(hessian.shape[0], dtype=bool)
-
-    covars = np.zeros_like(hessian)
-    for roi_ind in range(hessian.shape[0]):
-        if output_singularity:
-            is_singular[roi_ind] = True
-
-        try:
-            covars[roi_ind] = np.linalg.inv(hessian[roi_ind])
-
-            if output_singularity:
-                is_singular[roi_ind] = False
-        except np.linalg.linalg.LinAlgError:
-            try:
-                covars[roi_ind] = np.linalg.pinv(hessian[roi_ind])
-            except np.linalg.linalg.LinAlgError:
-                covars[roi_ind] = np.zeros(hessian[roi_ind].shape)
-
-    diagonal_ind = np.arange(hessian.shape[1])
-    covars[:, diagonal_ind, diagonal_ind] = np.abs(covars[:, diagonal_ind, diagonal_ind])
-
-    if output_singularity:
-        return covars, is_singular
-    return covars
+    if max_batch_size is not None:
+        offset = 0
+        elements_left = nmr_elements
+        while elements_left > 0:
+            batch_size = min(elements_left, max_batch_size)
+            yield offset, offset + batch_size
+            elements_left -= batch_size
+            offset += batch_size
+    else:
+        offset = 0
+        elements_left = nmr_elements
+        for ind in range(nmr_batches - 1):
+            batch_size = int(min(np.ceil(nmr_elements / nmr_batches), nmr_elements))
+            yield offset, offset + batch_size
+            elements_left -= batch_size
+            offset += batch_size
+        yield (offset, offset + elements_left)
 
 
 def covariance_to_correlations(covariance):
@@ -529,7 +633,7 @@ def multiprocess_mapping(func, iterable):
         func (func): the function to apply
         iterable (iterable): the iterable with the elements we want to apply the function on
     """
-    if os.name == 'nt': # In Windows there is no fork.
+    if os.name == 'nt':  # In Windows there is no fork.
         return list(map(func, iterable))
     try:
         p = multiprocessing.Pool()
@@ -541,38 +645,130 @@ def multiprocess_mapping(func, iterable):
         return list(map(func, iterable))
 
 
-_cl_functions_parser = tatsu.compile('''
-    result = {function}+;
-    function = [address_space] data_type function_name arguments body;
+_tatsu_cl_function = '''
+    function = {documentation}* [kernel] [address_space] data_type function_name arglist body;
+    documentation = '/*' ->'*/';
+    kernel = ['__'] 'kernel';
     address_space = ['__'] ('local' | 'global' | 'constant' | 'private');
     data_type = /\w+(\s*(\*)?)+/;
     function_name = /\w+/;
-    arguments = /\([\w,\*\s]+\)/;
-    body = compound_statement;    
+    arglist = '(' @+:arg {',' @+:arg}* ')' | '()';
+    arg = /[\w \*\[\]]+/;
+    body = compound_statement;
     compound_statement = '{' {[/[^\{\}]*/] [compound_statement]}* '}';
-''')
+'''
+
+_extract_cl_functions_parser = tatsu.compile('''
+    result = {function}+;
+''' + _tatsu_cl_function)
+
+_split_cl_function_parser = tatsu.compile('''
+    result = function;
+''' + _tatsu_cl_function)
 
 
-def separate_cl_functions(input_str):
-    """Separate all the OpenCL functions.
+def parse_cl_function(cl_code, dependencies=()):
+    """Parse the given OpenCL string to a single SimpleCLFunction.
 
-    This creates a list of strings, with for each function found the OpenCL code.
+    If the string contains more than one function, we will return only the last, with all the other added as a
+    dependency.
 
     Args:
-        input_str (str): the string containing one or more functions.
+        cl_code (str): the input string containing one or more functions.
+        dependencies (Iterable[CLCodeObject]): The list of CL libraries this function depends on
 
     Returns:
-        list: a list of strings, with one string per found CL function.
+        mot.lib.cl_function.SimpleCLFunction: the CL function for the last function in the given strings.
     """
-    class Semantics(object):
+    from mot.lib.cl_function import SimpleCLFunction
+
+    def separate_cl_functions(input_str):
+        """Separate all the OpenCL functions.
+
+        This creates a list of strings, with for each function found the OpenCL code.
+
+        Args:
+            input_str (str): the string containing one or more functions.
+
+        Returns:
+            list: a list of strings, with one string per found CL function.
+        """
+        class Semantics:
+
+            def __init__(self):
+                self._functions = []
+
+            def result(self, ast):
+                return self._functions
+
+            def arglist(self, ast):
+                if ast == '()':
+                    return '()'
+                return '({})'.format(', '.join(ast))
+
+            def function(self, ast):
+                def join(items):
+                    result = ''
+                    for item in items:
+                        if isinstance(item, str):
+                            result += item
+                        else:
+                            result += join(item)
+                    return result
+
+                self._functions.append(join(ast).strip())
+                return ast
+
+        return _extract_cl_functions_parser.parse(input_str, semantics=Semantics())
+
+    functions = separate_cl_functions(cl_code)
+    return SimpleCLFunction.from_string(functions[-1], dependencies=list(dependencies or []) + [
+        SimpleCLFunction.from_string(s) for s in functions[:-1]])
+
+
+def split_cl_function(cl_str):
+    """Split an CL function into a return type, function name, parameters list and the body.
+
+    Args:
+        cl_str (str): the CL code to parse and plit into components
+
+    Returns:
+        tuple: string elements for the return type, function name, parameter list and the body
+    """
+    class Semantics:
 
         def __init__(self):
-            self._functions = []
+            self._is_kernel_func = False
+            self._return_type = ''
+            self._function_name = ''
+            self._parameter_list = []
+            self._cl_body = ''
 
         def result(self, ast):
-            return self._functions
+            return self._is_kernel_func, self._return_type, self._function_name, self._parameter_list, self._cl_body
 
-        def function(self, ast):
+        def kernel(self, ast):
+            self._is_kernel_func = True
+            return ast
+
+        def address_space(self, ast):
+            self._return_type = ast.strip() + ' '
+            return ast
+
+        def data_type(self, ast):
+            self._return_type += ''.join(ast).strip()
+            return ast
+
+        def function_name(self, ast):
+            self._function_name = ast.strip()
+            return ast
+
+        def arglist(self, ast):
+            if ast != '()':
+                self._parameter_list = ast
+            return ast
+
+        def body(self, ast):
             def join(items):
                 result = ''
                 for item in items:
@@ -582,31 +778,8 @@ def separate_cl_functions(input_str):
                         result += join(item)
                 return result
 
-            self._functions.append(join(ast).strip())
+            self._cl_body = join(ast).strip()[1:-1]
             return ast
 
-    return _cl_functions_parser.parse(input_str, semantics=Semantics())
-
-
-def parse_cl_function(input_str, dependencies=(), cl_extra=None):
-    """Parse the given OpenCL string to a single SimpleCLFunction.
-
-    If the string contains more than one function, we will return only the last, with all the other added as a
-    dependency.
-
-    Args:
-        input_str (str): the input string containing one or more functions.
-        dependencies (list or tuple of CLLibrary): The list of CL libraries this function depends on
-        cl_extra (str): extra CL code for this function that does not warrant an own function.
-            This is prepended to the function body.
-
-    Returns:
-        mot.lib.cl_function.SimpleCLFunction: the CL function for the last function in the given strings.
-    """
-    from mot.lib.cl_function import SimpleCLFunction
-
-    functions = separate_cl_functions(input_str)
-    return SimpleCLFunction.from_string(functions[-1], dependencies=list(dependencies or []) + [
-        SimpleCLFunction.from_string(s) for s in functions[:-1]
-    ], cl_extra=cl_extra)
+    return _split_cl_function_parser.parse(cl_str, semantics=Semantics())
 
