@@ -4,7 +4,7 @@ __author__ = 'Robbert Harms'
 __date__ = "2014-02-05"
 __license__ = "LGPL v3"
 __maintainer__ = "Robbert Harms"
-__email__ = "robbert.harms@maastrichtuniversity.nl"
+__email__ = "robbert@xkls.nl"
 
 
 class MetropolisWithinGibbs(AbstractRWMSampler):
