@@ -7,7 +7,7 @@ from numpy.testing import assert_allclose
 __author__ = 'Robbert Harms'
 __date__ = '2018-05-12'
 __maintainer__ = 'Robbert Harms'
-__email__ = 'robbert.harms@maastrichtuniversity.nl'
+__email__ = 'robbert@xkls.nl'
 __licence__ = 'LGPL v3'
 
 
